"""Models Z-Wave devices."""

from .const import (
    _LOGGER,
)
from .device import GenericSwitch, GenericSwitchMultilevel, HomeSeerDevice


class InsteonSwitch(GenericSwitch):
    pass


class InsteonSwitchMultilevel(GenericSwitchMultilevel):
    pass


class InsteonSensorBinary(HomeSeerDevice):
    pass



"""ASCII/Telnet listener for HomeSeer."""

import asyncio

import logging

_LOGGER = logging.getLogger(__name__)

from .const import (
    REASON_RECONNECTED,
    STATE_IDLE,
    STATE_LISTENING,
    STATE_STOPPED,
)
from .errors import HomeSeerASCIIConnectionError


class ASCIIListener:
    def __init__(self, host, **kwargs):
        self._host = host
        self._port = kwargs.get("ascii_port", 11000)
        self._async_message_callback = kwargs.get("async_message_callback")
        self._async_disconnection_callback = kwargs.get("async_disconnection_callback")
        self._async_reconnection_callback = kwargs.get("async_connection_callback")
        self._username = kwargs.get("username")
        self._password = kwargs.get("password")
        self._reader = None
        self._writer = None
        self._reconnect_flag = False
        self._state = STATE_IDLE
        self._flag = True

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    async def _start_listener(self):
        connection = asyncio.open_connection(self._host, self._port)
        try:
            self._reader, self._writer = await asyncio.wait_for(connection, timeout=3)
            _LOGGER.info(
                f"HomeSeer ASCII Listener connected to {self._host}:{self._port}"
            )
            self._state = STATE_LISTENING

            if not await self._login():
                raise Exception("Error logging in to ASCII listener")

            self._flag = True
            asyncio.get_event_loop().create_task(self._pinger())

            if self._reconnect_flag and self._async_reconnection_callback is not None:
                await self._async_reconnection_callback(reason=REASON_RECONNECTED)
            self._reconnect_flag = False

            while True:
                msg = await self._reader.readline()
                _LOGGER.debug(f"HomeSeer raw ASCII message received: {msg}")
                if msg == b"":
                    raise HomeSeerASCIIConnectionError
                else:
                    await self._handle_message(msg.decode())

        except HomeSeerASCIIConnectionError:
            _LOGGER.error("Empty ASCII message received, connection error")
            await self._handle_disconnect()

        except asyncio.TimeoutError:
            _LOGGER.error("HomeSeer ASCII listener connect timed out")
            await self._handle_disconnect()

        except Exception as err:
            _LOGGER.error(f"HomeSeer ASCII listener error: {err}")
            await self._handle_disconnect()

    async def _login(self):
        if self._writer is None:
            return False

        auth = "au,{},{}\r\n".format(self._username, self._password).encode()
        _LOGGER.debug(
            f"HomeSeer ASCII logging in with user {self._username} and pass {self._password}"
        )
        self._writer.write(auth)

        try:
            msg = await asyncio.wait_for(self._reader.readline(), timeout=3)
            if msg.decode().strip() == "ok":
                _LOGGER.debug("HomeSeer ASCII login ok")
            else:
                _LOGGER.debug("HomeSeer ASCII login error: bad user or pass")
            return True
        except asyncio.TimeoutError:
            _LOGGER.error("HomeSeer ASCII login timeout")
            return False

    async def _handle_message(self, raw):
        msg = raw.split(",")
        self._flag = True
        if msg[0] == "DC" and self._async_message_callback is not None:
            await self._async_message_callback(msg[1], msg[2])
        else:
            _LOGGER.debug(
                f"HomeSeer unhandled ASCII message type received: {msg[0].strip()}"
            )

    async def _handle_disconnect(self):
        self._reconnect_flag = True
        if self._writer is not None:
            self._writer.close()

        if self._async_disconnection_callback is not None:
            await self._async_disconnection_callback()

        await self.connection_handler()

    async def _pinger(self):
        while True:
            if self.state == STATE_LISTENING:
                if self._flag:
                    self._flag = False
                    _LOGGER.debug("Sending ping...")
                    self._writer.write("vr\r\n".encode())
                    await self._writer.drain()
                    await asyncio.sleep(120)
                else:
                    _LOGGER.debug("Ping timeout, closing ASCII connection")
                    self._writer.close()
                    break
            else:
                break

    async def connection_handler(self):
        if self.state == STATE_STOPPED:
            _LOGGER.debug("Stopping and closing ASCII listener")
            self._reconnect_flag = True
            if self._writer is not None:
                self._writer.close()
        else:
            if self._reconnect_flag:
                time = 10
                _LOGGER.info(f"Reconnecting ASCII Listener in {time} seconds...")
                await asyncio.sleep(time)
            _LOGGER.debug(f"Connecting ASCII listener to {self._host}:{self._port}")
            self._state = STATE_IDLE
            asyncio.get_event_loop().create_task(self._start_listener())



from .zwave import (
    ZWaveBarrierOperator,
    ZWaveDoorLock,
    ZWaveSwitch,
    ZWaveSwitchMultilevel,
)
from .insteon import InsteonSwitch, InsteonSwitchMultilevel
import logging

_LOGGER = logging.getLogger(__name__)

from .device import (
    GenericBatterySensor,
    GenericBinarySensor,
    GenericEvent,
    GenericFanSensor,
    GenericHumiditySensor,
    GenericLuminanceSensor,
    GenericMultiLevelSensor,
    GenericOperatingStateSensor,
    GenericPowerSensor,
    GenericSensor,
    HomeSeerDevice,
)

# Key / Value
HASS_DICTIONARY = {
    # Insteon switchs
    "Insteon Dual-Band SwitchLinc On/Off": InsteonSwitch,
    "Insteon Outdoor ApplianceLinc": InsteonSwitch,
    "ApplianceLinc": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button B": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button C": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button D": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button E": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button F": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button G": InsteonSwitch,
    "Insteon KeypadLinc V2 Dimmer Button H": InsteonSwitch,
    "Insteon I/O Linc Output Relay": InsteonSwitch,
    # Insteon multilevel switch
    "Insteon Dual-Band SwitchLinc Dimmer": InsteonSwitchMultilevel,
    "Insteon KeypadLinc V2 Dimmer Button A/Load": InsteonSwitchMultilevel,
    # Insteon sensors
    "Insteon I/O Linc Input Sensor": GenericBinarySensor,
    # ZWave switch
    "Z-Wave Switch": ZWaveSwitch,
    "Z-Wave Switch Binary": ZWaveSwitch,
    # ZWave multilevel switch
    "Z-Wave Switch Multilevel": ZWaveSwitchMultilevel,
    # ZWave sensors
    "Z-Wave Sensor Binary": GenericBinarySensor,
    "Z-Wave Sensor Multilevel": GenericMultiLevelSensor,
    "Z-Wave Temperature": GenericMultiLevelSensor,
    "Z-Wave Battery": GenericBatterySensor,
    "Z-Wave Fan State": GenericFanSensor,
    "Z-Wave Relative Humidity": GenericHumiditySensor,
    "Z-Wave Operating State": GenericOperatingStateSensor,
    "Z-Wave Luminance": GenericLuminanceSensor,
    "Z-Wave Watts": GenericPowerSensor,
    "Z-Wave Electric Meter": GenericPowerSensor,
    # ZWave Locks
    "Z-Wave Door Lock": ZWaveDoorLock,
    # ZWave Cover
    "Z-Wave Barrier Operator": ZWaveBarrierOperator,
    # ZWave Events
    "Z-Wave Central Scene": GenericEvent,
    # DSC Sensors
    "BLDSC Plug-In Device": GenericBinarySensor,
}


def get_device(raw, control_data, request):
    device_type = raw["device_type_string"]

    # lookup
    try:
        type = HASS_DICTIONARY[device_type]
        return type(raw, control_data, request)
    except KeyError:
        return None

    return None


"""Models a HomeSeer event"""


class HomeSeerEvent:
    def __init__(self, raw, request):
        self._raw = raw
        self._request = request

    @property
    def group(self):
        return self._raw["Group"]

    @property
    def name(self):
        return self._raw["Name"]

    async def run(self):
        json = {"action": "runevent", "group": self.group, "name": self.name}
        await self._request("post", json=json)



"""Helpers for Home Assistant."""
from .const import (
    HS_UNIT_CELSIUS,
    HS_UNIT_FAHRENHEIT,
    HS_UNIT_LUX,
    HS_UNIT_PERCENTAGE,
    HS_UNIT_WATTS,
    HS_UNIT_KILOWATTS,
    HS_UNIT_AMPS,
    HS_UNIT_VOLTS,
)



"""
Models a connection to a HomeSeer HomeTroller (or independent HomeSeer HS3 software installation).
Allows sending commands via JSON API and listening for device changes via ASCII interface.
"""

from asyncio import TimeoutError
from .factory import get_device
from .device import HomeSeerDevice
from aiohttp import BasicAuth, ContentTypeError
from typing import Union

import logging

_LOGGER = logging.getLogger(__name__)

from .const import (
    DEFAULT_ASCII_PORT,
    DEFAULT_HTTP_PORT,
    DEFAULT_PASSWORD,
    DEFAULT_USERNAME,
    REASON_DISCONNECTED,
    STATE_IDLE,
    STATE_STOPPED,
)
from .events import HomeSeerEvent
from .listener import ASCIIListener


class HomeTroller:
    def __init__(
        self,
        host,
        websession,
        username=DEFAULT_USERNAME,
        password=DEFAULT_PASSWORD,
        http_port=DEFAULT_HTTP_PORT,
        ascii_port=DEFAULT_ASCII_PORT,
    ):
        self._host = host
        self._websession = websession
        self._auth = BasicAuth(username, password)
        self._http_port = http_port
        self._ascii_port = ascii_port
        self._listener = ASCIIListener(
            self._host,
            username=username,
            password=password,
            ascii_port=self._ascii_port,
            async_message_callback=self._update_device_value,
            async_connection_callback=self.refresh_devices,
            async_disconnection_callback=self._disconnect_callback,
        )
        self.devices = {}
        self.all_devices = {}
        self.events = []

    @property
    def state(self):
        return self._listener.state

    async def initialize(self):
        await self._get_devices()
        await self._get_events()

    async def start_listener(self):
        self._listener.state = STATE_IDLE
        await self._listener.connection_handler()

    async def stop_listener(self):
        self._listener.state = STATE_STOPPED
        await self._listener.connection_handler()

    async def control_device_by_value(self, ref: int, value: Union[str, int, float]):
        """
        Provides an interface for controlling devices by value
        directly through the HomeTroller object.
        """
        params = {
            "request": "controldevicebyvalue",
            "ref": ref,
            "value": value,
        }
        await self._request("get", params=params)

    async def _request(self, method, params=None, json=None):
        """Make an API request"""
        url = f"http://{self._host}:{self._http_port}/JSON"

        try:
            async with self._websession.request(
                method,
                url,
                params=params,
                json=json,
                auth=self._auth,
            ) as result:
                result.raise_for_status()
                _LOGGER.debug(f"HomeSeer request response: {await result.text()}")
                return await result.json()

        except ContentTypeError:
            _LOGGER.debug(f"HomeSeer returned non-JSON response: {await result.text()}")

        except TimeoutError:
            _LOGGER.error(
                f"Timeout while requesting HomeSeer data from {self._host}:{self._http_port}"
            )

        except Exception as err:
            _LOGGER.error(f"HomeSeer HTTP Request error: {err}")

    async def _get_devices(self):
        try:
            params = {"request": "getstatus"}
            result = await self._request("get", params=params)

            all_devices = result["Devices"]

            params = {"request": "getcontrol"}
            result = await self._request("get", params=params)

            control_data = result["Devices"]

            for device in all_devices:
                dev = get_device(device, control_data, self._request)

                if dev is not None:
                    self.devices[dev.ref] = dev
                else:
                    # Create a dummy device
                    dev = HomeSeerDevice(device, control_data, self._request)
                    _LOGGER.debug(
                        f"HomeSeer device type not supported: {dev.device_type_string}"
                    )
                # add to the all device list
                self.all_devices[dev.ref] = dev
        except TypeError:
            _LOGGER.error("Error retrieving HomeSeer devices!")

    async def _get_events(self):
        try:
            params = {"request": "getevents"}
            result = await self._request("get", params=params)

            all_events = result["Events"]

            for event in all_events:
                ev = HomeSeerEvent(event, self._request)
                self.events.append(ev)

        except TypeError:
            _LOGGER.error("Error retrieving HomeSeer events!")

    async def _update_device_value(self, device_ref, value):
        try:
            device = self.devices[int(device_ref)]
            prev_value = device.value
            device.update_value(value)
            _LOGGER.info(
                f"{device.device_type_string} '{device.name}' ({device.ref}) updated from: {prev_value} to: {device.value}"
            )

        except KeyError:
            try:
                device = self.all_devices[int(device_ref)]
                device.update_value(value)
                _LOGGER.debug(
                    f"Unsupported {device.device_type_string} '{device.name}' ({device.ref}) updated to: {device.value}"
                )

            except KeyError:
                _LOGGER.error(
                    f"HomeSeer update received for unknown device: {device_ref}"
                )

    async def _disconnect_callback(self):
        for device in self.devices.values():
            device.update_value(None, REASON_DISCONNECTED)

    async def refresh_devices(self, reason=None):
        try:
            params = {"request": "getstatus"}
            result = await self._request("get", params=params)

            all_devices = result["Devices"]

        except TypeError:
            _LOGGER.error("Error retrieving HomeSeer data for refresh!")
            return

        for device in all_devices:
            try:
                dev = self.devices[device["ref"]]
                dev.update_value(device["value"], reason)
                _LOGGER.debug(
                    f"HomeSeer device '{dev.name}' ({dev.ref}) refreshed to: {dev.value}"
                )
            except KeyError:
                _LOGGER.debug(
                    f"HomeSeer refresh data retrieved for unsupported device type: "
                    f"{device['device_type_string']} ({device['ref']})"
                )



import setuptools

long_description = open("README.md").read()

setuptools.setup(
    name="pyhs3ng",
    version="1.0.3",
    author="Jason Rhubottom",
    author_email="jason@rhusoft.com",
    description="Python3 async library for interacting with HomeSeer HS3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jrhubott/pyhs3",
    packages=["pyhs3ng"],
    install_requires=["asyncio", "aiohttp"],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
